package edu.princeton.swing;

import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.beans.*;
import java.util.Vector;

import javax.swing.*;
import javax.swing.border.*;

/**
 * PList extends JList to provide ClipboardTarget support.
 *
 * @author btsang
 * @version 7.1
 */

public class PList extends JList implements ClipboardTarget
{
    private static final Border EMPTY_BORDER = new EmptyBorder(1, 1, 1, 1);
    private static final Toolkit TOOLKIT = Toolkit.getDefaultToolkit();

    private JPopupMenu popupMenu;
    private Vector actionListeners;
    private String actionCommand;
    private Listener listener;
    protected Border focusedCellBorder;

    /**
     * Constructs a new PList.
     */
    public PList()
    {
        super();

        initList();
    }

    /**
     * Constructs a new PList.
     */
    public PList(ListModel dataModel)
    {
        super(dataModel);

        initList();
    }

    /**
     * Constructs a new PList.
     */
    public PList(Object[] listData)
    {
        super(listData);

        initList();
    }

    /**
     * Constructs a new PList.
     */
    public PList(Vector listData)
    {
        super(listData);

        initList();
    }

    /**
     * Called by the constructors to init the PList properly.
     */
    private void initList()
    {
        listener = new Listener();

        PropertiedListSelectionModel model = (PropertiedListSelectionModel)getSelectionModel();
        model.addPropertyChangeListener(listener);

        popupMenu = ClipboardTargetManager.createDefaultPopupMenu();
        setCellRenderer(new PListCellRenderer());

        actionListeners = new Vector();
        actionCommand = null;

        ClipboardTargetManager.installMappings(this);

        enableEvents(AWTEvent.FOCUS_EVENT_MASK | AWTEvent.MOUSE_EVENT_MASK);
    }

    /**
     * Adds a listener to monitor action events (enter key events) on the PTextField.
     *
     * @param listener The listener to add to this document.  A null value will cause a
     * NullPointerException.
     */
    public synchronized void addActionListener(ActionListener listener)
    {
        if (listener == null)
            throw new NullPointerException();

        actionListeners.add(listener);
    }

    /**
     * Removes a listener from this component.  Nothing will happen if no match was found.
     *
     * @param listener The listener to remove from this document.  A null value will cause a
     * NullPointerException.
     */
    public synchronized void removeActionListener(ActionListener listener)
    {
        if (listener == null)
            throw new NullPointerException();

        actionListeners.remove(listener);
    }

    /**
     * Returns the action command associated with the enter key being pressed.  A null value
     * indicates thatthe text of the text field at the time of the event will be used.
     *
     * return The action command to be associated with the enter key.
     */
    public String getActionCommand()
    {
        return actionCommand;
    }

    /**
     * Sets the action command associated with the enter key being pressed.
     *
     * @param actionCommand The action command to be associated with the enter key.  If null is
     * passed, the text of the text field at the time of the event will be used.
     */
    public void setActionCommand(String actionCommand)
    {
        if (this.actionCommand != actionCommand && 
            (actionCommand != null || !actionCommand.equals(this.actionCommand)))
        {
            String oldValue = this.actionCommand;
            this.actionCommand = actionCommand;

            firePropertyChange("actionCommand", oldValue, actionCommand);
        }
    }

    /**
     * Returns wheter or not the cut operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to cut() should be permitted.
     */
    public boolean canCut()
    {
        return false;
    }

    /**
     * Copies the selection to the clipboard and remove it from the object.  The clipboard should
     * not be modified if nothing is selected.
     */
    public void cut()
    {

    }

    /**
     * Returns wheter or not the copy operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to copy() should be permitted.
     */
    public boolean canCopy()
    {
        return isEnabled();
    }

    /**
     * Copies the selection to the clipboard.  The clipboard should not be modified if nothing is
     * selected.
     */
    public void copy()
    {
        if (isEnabled())
        {
            Object values[] = getSelectedValues();
            int length = values.length;
            if (length == 0)
                return;

            StringBuffer buffer = new StringBuffer();

            for (int ctr = 0; ctr < length; ctr++)
            {
                buffer.append(values[ctr]);
                if (ctr + 1 < length)
                    buffer.append('\n');
            }

            try
            {
                Clipboard clipboard = TOOLKIT.getSystemClipboard();

                StringSelection selection = new StringSelection(
                    buffer.toString()
                    );

                clipboard.setContents(selection, selection);
            }
            catch (Exception e)
            {

            }
        }
    }

    /**
     * Returns wheter or not the paste operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to paste() should be permitted.
     */
    public boolean canPaste()
    {
        return false;
    }

    /**
     * Inserts the contents of the clipbaord into the object.  <b>Optional</b>: Before inserting
     * the contents of the clipboard, remove the current selection from the object.
     */
    public void paste()
    {

    }

    /**
     * Returns wheter or not the selectAll operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to selectAll() should be permitted.
     */
    public boolean canSelectAll()
    {
        return isEnabled() && getSelectionMode() != ListSelectionModel.SINGLE_SELECTION;
    }

    /**
     * Selects everything in the object.
     */
    public void selectAll()
    {
        if (isEnabled() && getSelectionMode() != ListSelectionModel.SINGLE_SELECTION)
        {
            setSelectionInterval(0, getModel().getSize() - 1);
        }
    }

    /**
     * Intercept FocusEvents to notify the ClipboardTargetManager of the event.
     */
    protected void processFocusEvent(FocusEvent e)
    {
        if (e.getID() == FocusEvent.FOCUS_GAINED)
        {
            ClipboardTargetManager.targetGainedFocus(this);
        }
        else if (e.getID() == FocusEvent.FOCUS_LOST)
        {
            ClipboardTargetManager.targetLostFocus(this);
        }

        super.processFocusEvent(e);
    }

    /**
     * Intercept MouseEvents to activate the popupMenu at the right time.
     */
    protected void processMouseEvent(MouseEvent e)
    {
        if (!isEnabled())
            return;

        boolean passedToSuper = false;

        if (popupMenu != null && e.isPopupTrigger())
        {
            if (!hasFocus())
                requestFocus();

            int x = e.getX();
            int y = e.getY();
            Point componentLocation = getLocationOnScreen();
            Dimension screenSize = TOOLKIT.getScreenSize();
            Dimension preferredSize = popupMenu.getPreferredSize();

            if (componentLocation.x + x + preferredSize.width > screenSize.width)
                x -= preferredSize.width;

            if (componentLocation.y + y + preferredSize.height > screenSize.height)
                y -= preferredSize.height;

            e.consume();
            passedToSuper = true;

            super.processMouseEvent(e);

            popupMenu.show(this, x, y);
        }
        else if (e.getID() == MouseEvent.MOUSE_CLICKED && e.getClickCount() == 2 &&
                 (e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0)
        {
            // Dispatch the events
            ActionEvent actionEvent;
            String actionCommand = this.actionCommand;
            if (actionCommand == null)
            {
                Object value = getSelectedValue();
                if (value == null)
                    actionCommand = "";
                else
                    actionCommand = value.toString();
            }
            
            actionEvent = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, actionCommand);

            Object array[] = actionListeners.toArray();
            for (int ctr = 0; ctr < array.length; ctr++)
                ((ActionListener)array[ctr]).actionPerformed(actionEvent);
        }

        if (!passedToSuper)
            super.processMouseEvent(e);
    }

    /**
     * Override the setSelectionBackground() method to intercept any changes concerning the
     * background color of the selection.
     */
    public void setSelectionBackground(Color selectionBackground)
    {
        focusedCellBorder = new LineBorder(selectionBackground.darker(), 1);

        super.setSelectionBackground(selectionBackground);

        ClipboardTargetManager.targetAbilityChanged(this);
    }

    /**
     * Return the border that should be rendered for a focused cell.
     *
     * @return The border that should be rendered for a focused cell.
     */
    public Border getFocusedCellBorder()
    {
        return focusedCellBorder;
    }

    /**
     * Return the border that should be rendered for an unfocused cell.
     *
     * @return The border that should be rendered for an unfocused cell.
     */
    public Border getUnfocusedCellBorder()
    {
        return EMPTY_BORDER;
    }

    /**
     * Override the setEnabled() method to intercept any changes concerning the enabled status of
     * the component.
     */
    public void setEnabled(boolean enabled)
    {
        super.setEnabled(enabled);

        ClipboardTargetManager.targetAbilityChanged(this);
    }

    /**
     * Override the setSelectionModel method to reject any attempt to change the selection model
     * to a model which does not implement PropertiedListSelectionModel.
     */
    public void setSelectionModel(ListSelectionModel selectionModel)
    {
        if (!(selectionModel instanceof PropertiedListSelectionModel))
            throw new IllegalArgumentException();

        PropertiedListSelectionModel model = (PropertiedListSelectionModel)getSelectionModel();
        model.removePropertyChangeListener(listener);

        super.setSelectionModel(selectionModel);

        model = (PropertiedListSelectionModel)getSelectionModel();
        model.addPropertyChangeListener(listener);
    }


    /**
     * Override the createSelectionModel function to use our own selectionModel function which
     * actually monitors the selection mode.
     */
    protected ListSelectionModel createSelectionModel()
    {
        return new DefaultPropertiedListSelectionModel();
    }

    /**
     * Gets the popup menu for this component.
     *
     * @return The popup menu for the component.  Null is returned if this component does not have
     * a popup menu.
     */
    public JPopupMenu getPopupMenu()
    {
        return popupMenu;
    }

    /**
     * Sets the popup menu for this component.
     *
     * @param popupMenu The popup menu for the component.  A null value indicates that this
     * component should not have a popup menu.
     */
    public void setPopupMenu(JPopupMenu popupMenu)
    {
        if (popupMenu != this.popupMenu)
        {
            if (popupMenu != null && this.popupMenu.isVisible())
            {
                this.popupMenu.setVisible(false);
            }

            JPopupMenu oldValue = this.popupMenu;
            this.popupMenu = popupMenu;

            firePropertyChange("popupMenu", oldValue, popupMenu);
        }
    }

    /**
     * The Listener listenes to the selection model of this list to capture PropertyChangeEvents.
     *
     * @author btsang
     * @version 7.1
     */
    protected class Listener implements PropertyChangeListener
    {
        /**
         * Creates a new Listener.
         */
        protected Listener()
        {

        }

        /**
         * Implement PropertyChangeListener to notify the ClipboardTargetManager of a possible
         * change in the canX() methods.
         */
        public void propertyChange(PropertyChangeEvent e)
        {
            if ("selectionMode".equals(e.getPropertyName()))
                ClipboardTargetManager.targetAbilityChanged(PList.this);
        }
    }

    /**
     * PListCellRenderer is a subclass of DefaultListCellRenderer which will correctly
     * render the border specified by the PList.
     *
     * @author btsang
     * @version 7.1
     */
    protected static class PListCellRenderer extends DefaultListCellRenderer
    {
        /**
         * Create a new PListCellRenderer.
         */
        protected PListCellRenderer()
        {
            super();
        }

        /**
         * Override getListCellRendererComponent() to correctly render the border specified by
         * the PList.
         */
        public Component getListCellRendererComponent(JList list, Object value, int index,
                                                      boolean isSelected, boolean cellHasFocus)
        {
            Component comp = super.getListCellRendererComponent(
                list,
                value,
                index,
                isSelected,
                cellHasFocus
                );

            // Override the border if the cell has focus
            if (list instanceof PList)
            {
                if (cellHasFocus)
                    setBorder(((PList)list).getFocusedCellBorder());
                else
                    setBorder(((PList)list).getUnfocusedCellBorder());
            }
            else {
                if (cellHasFocus)
                    setBorder(UIManager.getBorder("List.focusCellHighlightBorder"));
                else
                    setBorder(EMPTY_BORDER);
            }

            return comp;
        }
    }
}